#! /bin/env python3
import os
import sys
import time
import random as rnd
import cv2
import numpy as np
import pickle as pickle
import game
import random
import MultiNEAT as NEAT
from MultiNEAT import EvaluateGenomeList_Serial
from MultiNEAT import GetGenomeList, ZipFitness

from concurrent.futures import ProcessPoolExecutor, as_completed

def build_input(g):
    out = []
    for y in range(8):
        for x in range(8):
            tmp = g.field[y][x]
            try:
                b = bin(tmp).replace("0b", "")
                if len(b) == 1:
                    out += [0.0, 0.0, float(b)]
                elif len(b) == 2:
                    out += [0.0, float(b[0]), float(b[1])]
                else:
                    out += [float(b[0]), float(b[1]), float(b[2])]
            except:
                out += [0.0, 0.0, 0.0]
    out += [random.randint(0, 100)/100.0,1.0]
    return out


def evaluate(genome):
    net = NEAT.NeuralNetwork()
    genome.BuildPhenotype(net)

    f = 0.0

    for i in range(10):
        g = game.GameField(file="AI1.txt", AI=True)

        while len(g.search_for_moves()) != 0:
            # do stuff and return the fitness
            net.Flush()
            net.Input(np.array(build_input(g))) # can input numpy arrays, too
                                              # for some reason only np.float64 is supported
            net.Activate()
            move = net.Output()
            print(move)
            pos1, pos2 = ((int(str(round(move[0])) + str(round(move[1])) + str(round(move[2])), 2), int(str(round(move[3])) + str(round(move[4])) + str(round(move[5])), 2)),
                          (int(str(round(move[6])) + str(round(move[7])) + str(round(move[8])), 2), int(str(round(move[9])) + str(round(move[10])) + str(round(move[11])), 2)))
            if game.match_move(g.search_for_moves(), (pos1, pos2)):
                g.move((pos1, pos2))
            else:
                f -= 1.0
        f += g.points
    return f / 10



params = NEAT.Parameters()
params.PopulationSize = 150
params.DynamicCompatibility = True
params.WeightDiffCoeff = 4.0
params.CompatTreshold = 2.0
params.YoungAgeTreshold = 15
params.SpeciesMaxStagnation = 15
params.OldAgeTreshold = 35
params.MinSpecies = 5
params.MaxSpecies = 10
params.RouletteWheelSelection = False
params.RecurrentProb = 0.0
params.OverallMutationRate = 0.8

params.MutateWeightsProb = 0.90

params.WeightMutationMaxPower = 2.5
params.WeightReplacementMaxPower = 5.0
params.MutateWeightsSevereProb = 0.5
params.WeightMutationRate = 0.25

params.MaxWeight = 8

params.MutateAddNeuronProb = 0.03
params.MutateAddLinkProb = 0.05
params.MutateRemLinkProb = 0.0

params.MinActivationA  = 4.9
params.MaxActivationA  = 4.9

params.ActivationFunction_SignedSigmoid_Prob = 0.0
params.ActivationFunction_UnsignedSigmoid_Prob = 1.0
params.ActivationFunction_Tanh_Prob = 0.0
params.ActivationFunction_SignedStep_Prob = 0.0

params.CrossoverRate = 0.75  # mutate only 0.25
params.MultipointCrossoverRate = 0.4
params.SurvivalRate = 0.2


def getbest(i):

    g = NEAT.Genome(0, 194, 0, 12, False, NEAT.ActivationFunction.UNSIGNED_SIGMOID, NEAT.ActivationFunction.UNSIGNED_SIGMOID, 0, params)
    pop = NEAT.Population(g, params, True, 1.0, i)
    pop.RNG.Seed(i)

    generations = 0
    for generation in range(1000):
        genome_list = NEAT.GetGenomeList(pop)
        fitness_list = EvaluateGenomeList_Serial(genome_list, evaluate, display=False)
        NEAT.ZipFitness(genome_list, fitness_list)
        pop.Epoch()
        generations = generation
    bests = [(x, x.GetLeader().GetFitness()) for x in pop.Species]
    bests.sort()

    return bests[0]

best = getbest(random.randint(0, 100))

with open("genome", "wb") as f:
    pickle.dump(best, f)
